# copyright 1999 McMillan Enterprises
# license: use as you please. No warranty.
#
# Utility to create a very simple config file
# and build a installer / executable for a script.
#
# If your script requires anything outside the *.py, *.pyd and *.dll
# dependencies, you'll need to build the config file yourself.
#
# Oh, yes. This does support Tk automatically.
#
# Changed by Thomas Heller: Does not overwrite existing file
# without confirmation.

import os
import sys
import string

tmplt = """\
[MYINSTALL]
type= INSTALL
name= Install_%(basename)s.exe
bindepends= %(basename)s.py
zlib = INSTALLZLIB
misc= %(misc)s
debug = %(debug)d
excludes = PyWinTypes20.dll, win32api

[MYSTANDALONE]
type= STANDALONE
name= %(basename)s.exe
script= %(scripts)s
zlib = APPZLIB
userunw = %(tk)d
support = 0
debug = %(debug)d

[APPZLIB]
name= %(basename)s.pyz
dependencies= %(basename)s.py
excludes= dospath, posixpath, macpath

[INSTALLZLIB]
name = installzlib.pyz
dependencies = installutils
includes = installutils
excludes = dospath, posixpath, macpath
"""

def create(sname, home, tk, debug):
    workdir, script = os.path.split(sname)
    basenm, ext = os.path.splitext(script)
    cfgfile = os.path.join(workdir, basenm + '.cfg')
    if os.path.exists (cfgfile):
        answer = ''
        while answer not in ('y', 'Y', 'n', 'N'):
            answer = raw_input ("File '%s' already exists. Overwrite? [yn]" % cfgfile)
            if answer in ('n', 'N'):
                sys.exit (1)
    misc = 'MYSTANDALONE'
    scripts = script
    if tk:
        misc = 'MYSTANDALONE, TK'
        scripts = 'runTK.py, %s' % script
    cfg = open(cfgfile, 'w')
    cfg.write(tmplt % {'basename':basenm,
                       'pyinsthome':home,
                       'misc':misc,
                       'scripts':scripts,
                       'tk':tk,
                       'debug':debug})
    cfg.close()
    tkcfg = ''
    if tk:
        tkcfg = os.path.join(home, 'projects', 'buildTK', 'tk.cfg')
    cmd = '%s %s %s %s' % (q(sys.executable),
                           q(os.path.join(home, 'Builder.py')),
                           q(cfgfile), q(tkcfg))
    print '>',cmd
    os.system(cmd)

def q(s):
    if string.find(s, ' ') > -1:
        return '"'+s+'"'
    return s
  
if __name__ == '__main__':
    debug = tk = 0
    if len(sys.argv) < 2:
        print "Usage: python %s <scriptname>.py [options]" % sys.argv[0]
        print " options = [-tk]"
        print "      tk -> install a private TCL/TK for use by the app"
    else:
        home = os.path.dirname(sys.argv[0])
        if home == '':
            home = '.'
        home = os.path.abspath(home)
        if "-tk" in sys.argv:
            tk = 1
        if "-d" in sys.argv:
            debug = 1
        create(sys.argv[1], home, tk, debug)
            
