#include <stdio.h>
#include <io.h>
#include <stdlib.h>
#include <windows.h>
#include "zlib.h"
#define MAGIC "MEI\014\013\012\013\015"	
#ifdef _CONSOLE
# define FATALERROR(x) printf(x)
# define OTHERERROR(x) printf(x)
#else
# define FATALERROR(x) MessageBox(NULL, x, "Fatal Error!", MB_OK | MB_ICONEXCLAMATION)
# define OTHERERROR(x) MessageBox(NULL, x, "Error!", MB_OK | MB_ICONWARNING)
#endif

#ifdef _DEBUG
# define VS(arg) fprintf(stderr, arg)
# define VS1(fmt,arg) fprintf(stderr,fmt,arg)
# define VS2(fmt,arg1,arg2) fprintf(stderr,fmt,arg1,arg2)
#else
# define VS(arg)
# define VS1(fmt,arg)
# define VS2(fmt,arg1,arg2)
#endif

struct TOC {
	int structlen;	//len of this one - including full len of name
	int pos;	//pos rel to start of concatenation
	int len;	//len of the data (compressed)
	int ulen;	//len of data (uncompressed)
	char cflag;	//is it compressed (really a byte)
	char typcd; // 'b' binary, 'z' zlib, 'm' module, 's' script - added v3
	char name[1];	//the name to save it as
};
struct COOKIE {
	char magic[8]; 	// 'MEI\014\013\012\013\015'
	int  len;	//len of entire package
	int  TOC;	//pos (rel to start) of TableOfContents
	int  TOClen;	//length of TableOfContents
};

typedef void *PyObject;
#define Py_file_input 257

#define GETPROC(result, name, args)\
    typedef result (__cdecl *__PROC__##name) args;\
    __PROC__##name name = (__PROC__##name)GetProcAddress (dll, #name);\
    if (!name)\
        FATALERROR ("Cannot GetProcAddress for " #name)

#define DECLPROC(result, name, args)\
    typedef result (__cdecl *__PROC__##name) args;\
    __PROC__##name name = NULL;

#define GETGPROC(result, name, args)\
    name = (__PROC__##name)GetProcAddress (dll, #name);\
    if (!name)\
        FATALERROR ("Cannot GetProcAddress for " #name)

DECLPROC (PyObject *, Py_CompileString, (char *, char *, int));
DECLPROC (PyObject *, PyImport_ExecCodeModule, (char *, PyObject *));
DECLPROC (int, PyRun_SimpleString, (char *));

DECLPROC (PyObject *, PyErr_Occurred, () );
DECLPROC (void, PyErr_Print, () );

#define GETVAR(vartyp, name)\
	vartyp *name = (vartyp *)GetProcAddress (dll, #name);

//Forward
unsigned char *extract(struct TOC *toc, FILE *fp, int pkgstart);
int extract2fs(struct TOC *toc, FILE *fp, int pkgstart, char *homepath, char *temppath);
void import(struct TOC *toc, FILE *fp, int pkgstart);
void install(struct TOC *toc, FILE *fp, int pkgstart, char *arname);
void run(struct TOC *toc, FILE *fp, int pkgstart);
typedef int (__cdecl *PYMAIN)(int, char **);
void exitFunc();
//Globals
char *TOCbuff = NULL;
char homepath[_MAX_PATH+1];
char temppath[_MAX_PATH+1];
struct COOKIE c;
HINSTANCE dll;

#ifdef _CONSOLE
int main(int argc, char* argv[])
#else
APIENTRY WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
						LPSTR lpCmdLine, int nCmdShow )
#endif
{
	VS("Start\n");
	atexit( exitFunc );
	char thisfile[_MAX_PATH+1];
	if (!GetModuleFileNameA(NULL, thisfile, _MAX_PATH)) {
		FATALERROR("System error - unable to load!");
		return -1;
	}
	char *p = thisfile+strlen(thisfile) - 4;
	if (strnicmp(p, ".exe", 4) != 0)
		strcat(thisfile, ".exe");

	//GetModuleFileName returns an absolute path
	strcpy(homepath, thisfile);
	for (p=homepath+strlen(homepath); *p != '\\' && p >= homepath; --p);
	*++p = '\0';

	//Get path to official temp directory
	
	GetTempPath(MAX_PATH, temppath);

	//Check if homepath is read-only
	//if ((GetFileAttributes(thisfile) & FILE_ATTRIBUTE_READONLY)
	//	|| (access(thisfile, 2) != 0))
	//	strcpy(homepath, temppath);

	VS("Opening self\n");
	FILE *fp = fopen(thisfile, "rb");
	if (fp == NULL) {
		FATALERROR("Cannot open self: ");
		FATALERROR(thisfile);
		return -1;
	}
	fseek(fp, 0, SEEK_END);
	int filelen = ftell(fp);
	if (fseek(fp, -(int)sizeof(COOKIE), SEEK_END)) {
		FATALERROR("Archive appears to be invalid");
		return -1;
	}
	fread(&c, sizeof(COOKIE), 1, fp);
	if (strncmp(c.magic, MAGIC, 8)) {
		FATALERROR("Archive has bad magic - quiting!");
		return -1;
	}
	int pkgstart = filelen - c.len;
	fseek(fp, pkgstart+c.TOC, SEEK_SET);
	TOCbuff = (char *)malloc(c.TOClen);
	if (TOCbuff == NULL) {
		FATALERROR("Could not allocate buffer for TableOfContents\n");
		return -1;
	}
	fread(TOCbuff, c.TOClen, 1, fp);
	struct TOC *toc = (struct TOC *)TOCbuff;
	//pass 1 - unpack all the binary members (python20.dll, zlib.pyd etc...)
	VS("Unpacking binaries\n");
	while ((char*)toc < TOCbuff + c.TOClen) {
		if (toc->typcd == 'b')
			if (!extract2fs(toc, fp, pkgstart, homepath, temppath)) 
				FATALERROR("Could not extract binaries.\n");
		toc = (struct TOC *)((char*)toc + toc->structlen);
		if (toc == NULL) {
			FATALERROR("Cannot read Table of Contents.\n");
			return -1;
		}
	}
	int rc = 0;
	char **parg;
	//now launch python ...
	char *pypath = (char*)malloc(strlen(homepath)+strlen(temppath)+13);
	strcpy(pypath, "PYTHONPATH=");
	strcat(pypath, homepath);
	strcat(pypath, ";");
	strcat(pypath, temppath);
	pypath[strlen(pypath)-1] = '\0';
	VS("Manipulating evironment\n");
	_putenv(pypath);
	_putenv ("PYTHONHOME=");
#  ifdef _CONSOLE
	parg = argv;
#  else
	parg = __argv;
#  endif
	VS("Loading python\n");
	char *dllpath = (char*)malloc(max(strlen(homepath), strlen(temppath))+14);
	strcpy(dllpath, homepath);
	strcat(dllpath, "python20.dll");
	dllpath[strlen(dllpath)] = '\0';
	VS(dllpath);
	VS("\n");
	dll = LoadLibraryEx(dllpath, NULL, LOAD_WITH_ALTERED_SEARCH_PATH );	
	if (dll == 0) {
		strcpy(dllpath, temppath);
		strcat(dllpath, "python20.dll");
		dllpath[strlen(dllpath)] = '\0';
		VS(dllpath);
		VS("\n");
		dll = LoadLibraryEx(dllpath, NULL, LOAD_WITH_ALTERED_SEARCH_PATH );	
	}

	if (dll > 0) {
		int i;
		int vflag_index = -1;
		// Thomas Heller's code
		typedef void *PyObject;
		#define Py_file_input 257
		VS("Getting entry points\n");
		GETVAR (int, Py_NoSiteFlag);
		GETVAR (int, Py_VerboseFlag);
		GETPROC (int, Py_Initialize, (void));
		GETPROC (int, Py_Finalize, (void));
		GETGPROC (PyObject *, Py_CompileString, (char *, char *, int));
		GETGPROC (PyObject *, PyImport_ExecCodeModule, (char *, PyObject *));
		GETGPROC (int, PyRun_SimpleString, (char *));
		GETPROC (int, PySys_SetArgv, (int, char **));
		GETPROC (void, Py_SetProgramName, (char *));
		GETPROC (PyObject *, PyImport_ImportModule, (char *));
		GETPROC (int, PyObject_SetAttrString, (PyObject *, char *, PyObject *));
		GETPROC (PyObject *, PyList_New, (int));
		GETPROC (int, PyList_Append, (PyObject *, PyObject *));
		GETPROC (PyObject *, Py_BuildValue, (char *, ...));
		GETPROC (PyObject *, PyFile_FromFile, (FILE *, char *, char *, int));
		GETPROC (PyObject *, PyObject_CallFunction, (PyObject *, char *, ...));
		GETPROC (PyObject *, PyModule_GetDict, (PyObject *));
		GETPROC (PyObject *, PyDict_GetItemString, (PyObject *, char *));

		GETPROC (PyObject *, PyErr_Occurred, () );
		GETPROC (void, PyErr_Print, () );

#ifndef _CONSOLE
#define argc __argc
#define argv __argv
#endif
		VS("Setting Python vars\n");
		*Py_NoSiteFlag = 1;

		for (i = 0; i < argc; ++i) {
			if (0 == strcmp (argv[i], "-v")) {
				*Py_VerboseFlag = 1;
				vflag_index = i;
			}
		}
		
		Py_SetProgramName (thisfile);
		VS("Initializing Python\n");
		Py_Initialize ();
		VS("Manipulating Python's sys.path\n");
		PyRun_SimpleString ("import sys; del sys.path[3:]");

		PyObject *py_argv = PyList_New (0);
		PyObject *val = Py_BuildValue ("s", thisfile);
		PyList_Append (py_argv, val);
		for (i = 1; i < argc; ++i) {
			if (i != vflag_index) {
				PyObject *val = Py_BuildValue ("s", argv[i]);
				PyList_Append (py_argv, val);
			}
		}
		PyObject *sys = PyImport_ImportModule ("sys");
		VS("Setting sys.argv\n");
		PyObject_SetAttrString (sys, "argv", py_argv);
#ifndef _CONSOLE
#undef argc
#undef argv
#endif
		// pass 2 - "import" all the modules
		VS("importing modules from CArchive\n");
		PyObject *marshal = PyImport_ImportModule("marshal");
		PyObject *marshaldict = PyModule_GetDict(marshal);
		PyObject *loadfunc = PyDict_GetItemString(marshaldict, "load");
		PyObject *pyfile = PyFile_FromFile(fp, thisfile, "rb", 0);
		toc = (struct TOC *)TOCbuff;
		char *q;
		char buff[40];
		char cmd[80];
		while ((char*)toc < TOCbuff + c.TOClen) {
			if (toc->typcd == 'm' || toc->typcd == 'M') {
#ifdef _DEBUG
				fprintf(stderr, "loading %s\n", toc->name);
#endif
				fseek(fp, pkgstart + toc->pos + 8, SEEK_SET);
				PyObject *co = PyObject_CallFunction(loadfunc, "O", pyfile);
				PyObject *mod = PyImport_ExecCodeModule (toc->name, co);
				if( mod == NULL )
					fprintf(stderr, "mod is NULL loading %s\n", toc->name );
				if( PyErr_Occurred() )
					{
					fprintf(stderr, "error occurred loading %s\n", toc->name );
					PyErr_Print();
					}

				if (q=strrchr(toc->name, '.')) {
					strncpy(buff, toc->name, q - toc->name);
					buff[q - toc->name] = '\0';
					sprintf(cmd, "sys.modules[\"%s\"]%s = sys.modules[\"%s\"]", buff, q, toc->name);
#ifdef _DEBUG
					fprintf(stderr, "Run_SimpleString( %s )\n", cmd);
#endif
					PyRun_SimpleString(cmd);
				}
			}
			toc = (struct TOC *)((char*)toc + toc->structlen);
			if (toc == NULL) {
				FATALERROR("Cannot read Table of Contents.\n");
				return -1;
			}
		}
		// pass 3 - install all the zlibs
		VS("Installing import hooks\n");
		toc = (struct TOC *)TOCbuff;
		while ((char*)toc < TOCbuff + c.TOClen) {
			if (toc->typcd == 'z') 
				install(toc, fp, pkgstart, thisfile);
			toc = (struct TOC *)((char*)toc + toc->structlen);
			if (toc == NULL) {
				FATALERROR("Cannot read Table of Contents.\n");
				return -1;
			}
		}
		// pass 4 - run any scripts
		VS("Running scripts\n");
		toc = (struct TOC *)TOCbuff;
		while ((char*)toc < TOCbuff + c.TOClen) {
			if (toc->typcd == 's') 
				run(toc, fp, pkgstart);
			toc = (struct TOC *)((char*)toc + toc->structlen);
			if (toc == NULL) {
				FATALERROR("Cannot read Table of Contents.\n");
				return -1;
			}
		}
		VS("Finalizing Python\n");
		Py_Finalize ();
	}

	free(pypath);
	VS("closing self\n");
	fclose(fp);
	VS1("Returning %d to OS\n", rc);
	return rc;
}
//forward
unsigned char *decompress(unsigned char *, struct TOC *);

int extract2fs(struct TOC *toc, FILE *fp, int pkgstart, char *homepath, char *temppath)
{
	char fnm[_MAX_PATH+1];
	unsigned char *data = extract(toc, fp, pkgstart);
	strcpy(fnm, homepath);
	strcat(fnm, toc->name);
	FILE *out = fopen(fnm, "wb");
	if (out == NULL) {
		strcpy(fnm, temppath);
		strcat(fnm, toc->name);
		out = fopen(fnm, "wb");
		if (out == NULL) {
			free(data);
			return 0;
		}
	}
	fwrite(data, toc->ulen, 1, out);
	fclose(out);
	free(data);
	return 1;
}

void install(struct TOC *toc, FILE *fp, int pkgstart, char *arname)
{
	int zlibpos = pkgstart + toc->pos;
	char *tmpl = "import archive_rt;sys.importers.append(archive_rt.ZlibArchive(r'%s', %d))\n";
	char *cmd = (char *)malloc (strlen (tmpl) + strlen (arname) + 32);
	sprintf (cmd, tmpl, arname, zlibpos);
#if _DEBUG
	fprintf(stderr, "executing %s\n", cmd);
#endif
	PyRun_SimpleString (cmd);
	free(cmd);
}
void run(struct TOC *toc, FILE *fp, int pkgstart)
{
	unsigned char *data = extract(toc, fp, pkgstart);
#ifdef _DEBUG
	fprintf(stderr, "running %s\n", toc->name);
#endif
	PyRun_SimpleString((char *)data);
	free(data);
}
unsigned char *extract(struct TOC *toc, FILE *fp, int pkgstart)
{
#if _DEBUG
	fprintf(stderr, " extracting %s (%d, %c)\n", toc->name, toc->cflag, toc->typcd);
#endif
	fseek(fp, pkgstart + toc->pos, SEEK_SET);
	unsigned char *data = (unsigned char *)malloc(toc->len);
	if (data == NULL) {
		OTHERERROR("Could not allocate read buffer\n");
		return NULL;
	}
	fread(data, toc->len, 1, fp);
	if (toc->cflag == '\1') {
		unsigned char *tmp = decompress(data, toc);
		free(data);
		data = tmp;
		if (data == NULL) {
			char msg[40];
			sprintf(msg, "Error decompressing %s\n", toc->name);
			OTHERERROR(msg);
			return NULL;
		}
	}
	return data;
}

unsigned char *decompress(unsigned char * buff, struct TOC *toc)
{
	const char *ver = (zlibVersion)();
	unsigned char *out = (unsigned char *)malloc(toc->ulen);
	if (out == NULL) {
		OTHERERROR("Error allocating decompression buffer\n");
		return NULL;
	}
	z_stream zstream;

	zstream.zalloc = NULL;
	zstream.zfree = NULL;
	zstream.opaque = NULL;
	zstream.next_in = buff;
	zstream.avail_in = toc->len;
	zstream.next_out = out;
	zstream.avail_out = toc->ulen;
	int rc = inflateInit(&zstream);
	if (rc >= 0) {
		rc = (inflate)(&zstream, Z_FINISH);
		if (rc >= 0) {
			rc = (inflateEnd)(&zstream);
		}
		else {
			char msg[40];
			sprintf(msg, "Error %d from inflate: %s\n", rc, zstream.msg);
			OTHERERROR(msg);
			return NULL;
		}
	}
	else {
		char msg[40];
		sprintf(msg, "Error %d from inflateInit: %s\n", rc, zstream.msg);
		OTHERERROR(msg);
		return NULL;
	}	
	return out;
}

static	int removeFromFS(struct TOC *toc,const char* rootPath)
{
	int	r;
	char name[_MAX_PATH+1];
	strcpy(name, rootPath);
	strcat(name, toc->name);
	HMODULE h = GetModuleHandle(name);
	if(h){
		int i=0;
		while(FreeLibrary(h) && i<100) i++;
		VS2("Library %s freed %d times\n",name,i); 
		}
	r = remove(name);
	VS2("File '%s' removed\n",name, r ? "not" : "");
	return r;
}

void exitFunc()
{
	struct TOC *toc = (struct TOC *)TOCbuff;
	if (toc == NULL) {
		VS("exitFunc: Cannot read Table of Contents!\n");
		}
	else{
		while((char*)toc < TOCbuff + c.TOClen){
			if (toc->typcd == 'b'){
				if (removeFromFS(toc,homepath)==-1){
					removeFromFS(toc,temppath);
					}
				}
			toc = (struct TOC *)((char *)toc + toc->structlen);
			if (toc == NULL) {
				VS("exitFunc: Table of Contents went bad!\n");
				}
			}
		}
	int tries = 20;
	int rc = 1;
	while (tries-- && rc) rc = FreeLibrary(dll);
	//free(TOCbuff);
}
